/*======================================================================*
 *		        						* 
 * TITLE:	MYACC -- LALR(1) parser generator for modula-2	        *
 *									*
 * FILE:	data.c - data manager module     			*
 * 					        			* 
 * UPDATED:     July 14, 1987						*
 *									*
 * AUTHOR:      Werner Stanglow						*
 *								        *
 *======================================================================*/

#include "mdefs.h"

extern lflag;

/* --------------------------------------------------------------------	*/
arout(s,v,n)		/* output of array v to fdata or ftab		*/
		
char *s;		/* name of table				*/
int *v, n;		/* put out first n elements of v		*/ 

{	register i;
	char name[50];			/* internal buffer */ 

	strcpy(name,s); 
	strcat(name,"last");		
	wdef(flocal,name,n-1);
	
	if (lflag)
	{	if (fwrite(v,sizeof(int),n,ftab) != n)
			error("cannot write %s to table file %s",s, FTAB);
		fprintf(fdata,
			"\n   AND yyread(SYSTEM.ADR(%s),SIZE(INTEGER),",s);
		fprintf(fdata,"%s+1)\n",name);
	}
	else 
		for (i = 0; i < n; ++i)
			fprintf (fdata, "\t%12s [%3d] := %6d;\n",s,i,v[i]);
}

/* --------------------------------------------------------------------	*/
outreadtok()

{
	if (lflag)
	{ 	fprintf(fdata,
			"   AND yyread (SYSTEM.ADR(yytokens),");
		fprintf(fdata,"SIZE(yytokensAT),1)\n");
	}
}

/* --------------------------------------------------------------------	*/
static outset(fp,v,n)
FILE *fp;
int v[],n;

{ 	register i,j,x;
	register fst = 1;
	for (i=0; i <=n-1; ++i)
	{	for (x=v[i], j=31; x && j>=0; x>>=1, --j)
		{ 	if (x & 01)
			{	if (!fst)
					putc(',',fp);
				else
					fst = 0;
				fprintf(fp," %d",i*32+j);	
			}
		}
		if (v[i])
			putc('\n',fp);	
	}
}
	
static index = 0;
static int yys = 0;
static int *yyexc = 0;
#define DEFSZ 128

/* --------------------------------------------------------------------	*/
addelem(val) 

int val;

{	
	if (lflag)
	{	if (!yyexc)
		{	extern char *malloc();
			if (!(yyexc = (int *) malloc(sizeof(int) * DEFSZ)))
				error("allocation failed");
			yys =  DEFSZ;
		}
		else if (index >=yys)
		{	yys *= 2;
			if (!(yyexc = (int *) realloc(yyexc,sizeof(int)*yys)))
				error("reallocation failed");
		}
		yyexc[index++] = val;
	}
	else
	{	fprintf(fdata,"\t%12s [%3d] := %6d;\n","yyexca",
			index++,val);
	}
}

/* --------------------------------------------------------------------	*/
endarray()

{	if (lflag)
		arout("yyexca",yyexc,index);
	else
		wdef(flocal,"yyexcalast",index-1);
}


/* --------------------------------------------------------------------	*/
wdef(fp,s,n)

FILE * fp;
char * s;
int n;

{	
	if ((fp != flocal) && (fp != fdef))
		fprintf(fp,"CONST\n");
	fprintf(fp,"   %-15s = %4d;\n",s,n<0?0:n);
	if (n<0)
		fprintf(fp,"\t\t(* was negative, set to 0!*)\n");
}

/*
static struct tn
	{	char * name;
		int val;
	}	toknames[128], *tp = toknames;
		


addtok(name,val)

char *name;		
int val;	

{	
	tp->name = name;
	tp->val  = val;
	tp++; 
}


*/
/* --------------------------------------------------------------------	*/
/*
gentokset(max)	
int max;			
{ 	struct tn * t;

	fprintf(stderr,"max is %d\n",max);
	for (t=toknames; t<tp; t++)
		extout(t->name,max);		
}
*/
/* --------------------------------------------------------------------	*/
static extout(name,len)		/* output a string with fixed length	*/

char *name;			/* put out				*/
int len;			/* extend string up to len by '\0'	*/

{	register i,j,l;

	j = strlen(name);
	l = len > j ? len + 1 : j + 1;
	l = l % 4 ? l + 4 - l % 4 : l;
	for (i=0; i < j; i++)
		putc(name[i],ftab);
	for (i=j; i< l; i++)
		putc('\0',ftab);
}

/* --------------------------------------------------------------------	*/
defout()			/* put out the token definitions	*/
{
	register int i,x;
	register int maxtoklen = 12;
	char any = 0;

	if ( 3 <= ntokens)	
		fprintf(fdef,"\n(* Token definitions *)\nCONST\n");
	else
		return;
	for (i = 3; i <= ntokens; ++i)
	{
		if (tokset[i].name[0] == ' ')
			continue;
		outtoken(tokset[i].name,maxtoklen,i-3,tokset[i].value);
		any = 1;
		{	x = strlen(tokset[i].name);
			maxtoklen = maxtoklen < x ? x : maxtoklen;
		}
	}
	if (x= (maxtoklen % sizeof(int)))
		maxtoklen += (4-sizeof(int));
	if (any)
		outreadtok();
	wdef(flocal,"yytokenslast",extval - 0400 - 2);
	wdef(flocal,"yytokstrlenlast",maxtoklen-1);
	wdef(flocal,"yytermlast",ntokens);
	for (i = 0; i <= ntokens; ++i)
	{	int val = (tokset[i].value)*(tokset[i].iskey? -1:1);
		if (!lflag)
			fprintf(fdata,"   yytoksym[%d] := %d;\n",i,val);
		else if ( fwrite (&val,sizeof(int),1,ftab)!=1)
			error("cannot write to fdata");
	}
	if (lflag)
	{	fprintf(fdata,
			"\n   AND yyread(SYSTEM.ADR(yytoksym),");
		fprintf(fdata,"SIZE(INTEGER),yytermlast+1)\n");
	}
}				
