/*======================================================================*
 *		        						* 
 * TITLE:	MYACC -- LALR(1) parser generator for modula-2	        *
 *									*
 * FILE:	lex.c - lexical analyzer module				*
 * 					        			* 
 * DESCRIPTION: This file contains myacc's scanner.			*
 * 	        Function gettok() recognizes the tokens defined in      *
 *		lex.h.						        *
 * 									*	
 * UPDATED:     July 5, 1987						*
 * 									*	
 * AUTHOR: 	Werner Stanglow						*
 *								        *
 *======================================================================*/

#include "mdefs.h"
#include "search.h"
#include "token.h"

#define TEXTMAX 512 

static char lltokname[NAMESIZE];
char *tokname = lltokname;
int numbval;

static char text[TEXTMAX];
static char *tp = text;

char * lltext = text; 
/* 
 * locals (define a buffer to put characters back into input stream finput)
 */

#define BUFSZ 300
static inbuf[BUFSZ];
static int *bp = inbuf;

#define unget(c) *bp++ =c,--tp 
static int get()	/* - get next char from input stream finput	*/
{
	register c;
	if ((c = (bp==inbuf?getc(finput):*--bp)) != EOF)
		*tp++ = c;
	return c;
}

gettok()		/* - get next token from input stream finput	*/
	  		/* - returns token values defined in token.h	*/
{	 
	register c;    
	tp = text;
	c = skip();
	switch (c)		/* skip spaces and comments */
	{	case EOF:
               		c = ENDFILE;
			break;
		case '%':
			c = getkeyword();
			break;
		case '\'':
		case '\"':
			c = getstring(c);
			break;
		default:
		  	if (isalpha(c))
				c = getidentifier(c); 
			else if (isdigit(c))
				c = getnumber(c);
			break;
	}
	*tp ='\0'; 
	return c;
}

/*
 * auxiliary functions
 * ===================
 */

static TABLE myacckeys[] = {
#include "key.h"
			   }; 

static getkeyword()

/* 
 * - find myacc's keywords (called if '%' was read)
 */

{	register c,i;
	
	switch (c = get())
	{
		case EOF:
			return ERRORTOK;
		case '%':
			return MARK;
		case '{':
			return LCURLY;
		case '}': 
			return RCURLY;
		default:
			lltokname[i=0] = c;
			while ( (c = get()) != EOF && isalpha(c) )
				if (i < (NAMESIZE-1))
					lltokname[++i] = c;
			lltokname[++i] = '\0';
			unget(c);
			
			if (i=bsearch(lltokname,myacckeys,sizeof(myacckeys)/ 
					sizeof(myacckeys[0])))
				return i;
			else
				return ERRORTOK; 
	}
}

/*
 * get an identifier ( letter {digit|letter} ) 
 * check if it is followed by a colon         
 */

static getidentifier(first)

	int first;

{	register c, i;
	lltokname[i=0] = first;
	while ( ((c = get()) != EOF) && isalnum(c))
		if (i < (NAMESIZE-1))
			lltokname[++i] = c;
	lltokname[++i] = '\0';
	unget(c);
	if ( (c = skip()) == ':')		/* look ahead for ':' */
		return C_IDENTIFIER;
	else
	{	unget(c);
		return IDENTIFIER;
	}
}

 
/*      
 * get a (modula) string constant and determine if it is a character
 * constant ('c' or "c")               
 * 
 */

static getstring(match)

	int match;

{	register c, i;
	lltokname[0] = ' ';
	i = 1;
	while ( ((c = get()) != EOF) && (c != match) && (c != '\n') )
		lltokname[i++] = c;
	lltokname[i] = '\0';
	if (c == EOF)
		error("Unexpected EOF in string constant");
	else if (c == '\n')
		error("Newline in string constant");
	else if (i==2)		/* c matches */
		return IDENTIFIER;
	else 
		return STRING;
}


/*
 * get an nonnegative integer ( called when a digit was read)
 */

static getnumber(c)

	register c;

{	register octval = c - '0';
	register isoct = (c <= '7');
	numbval = c - '0';
	while((c = get()) != EOF && isdigit(c))
	{	numbval = numbval * 10  + c - '0';
		octval  = octval * 8 + c - '0';
		if (c > '7')
		  	isoct = 0;   
	}	
	if ( c == 'C' || c == 'c')
	{	
		if (!isoct)
			error("character constant contains decimal digits");
		lltokname[0] = ' ';
		lltokname[1] = (char) octval;
		lltokname[2] = '\0';
		return IDENTIFIER; 
	}
	else
	{	unget(c);
		return NUMBER;
	}
}


static int skip()
   
/*     	- skips spaces and comments             
 *      - returns first following significant character
 *      - calls error() if EOF within comment is detected
 */

{	register c;

	for(;;)
	{	
		c = get();

		if ( c == EOF )
		/*if ( (c=get()) == EOF ) */
			return EOF;
		else if (isspace(c))
		{	if ( c=='\n')
				lineno++;
			continue;
		}
		else if (c == '(')
		{
			if ((c = get()) != '*')		  
			{
				unget(c);
				return '(';
			}
			else
			{	register offset = lineno;
				register opencom = 1;
				register char  begin = 0, end = 0;

				while (opencom)
					switch (c = get())
				 	{ 	case EOF:
							error(     
		"unexpected EOF within comment beginning at line %d\n", offset);
						case '\n':
							lineno++;	
							begin = end = 0; 
						case '*':
							if (begin)
								opencom++;      
							else
								end = 1; 
							begin = 0;
							break;
						case ')':
							if (end)
								opencom--;
							begin = end = 0;
							break;
						case '(':
							begin = 1;
							end = 0;
							break;
						default:
							begin = end = 0;
					}	
			continue; 
			}	
		}
		else
			return (c);
	
	}
}
