/*======================================================================*
 *		        						* 
 * TITLE:	MYACC -- LALR(1) parser generator for modula-2	        *
 *									*
 * FILE:	main.c - main module 					*
 *									*
 * UPDATED:     October 10,						*
 * 									*	
 * AUTHOR: 	Werner Stanglow						*
 *								        *
 *======================================================================*/

#define	GLOBAL	/**/
#include "mdefs.h"
#include "main.h"

#ifdef SIGAVAI
#include <signal.h>
#endif

#define	USAGE {fprintf(stderr,\
	"myacc [-vrelkKtbl] [-o modname] [-W dir] grammar\n");\
	       exit(EX_SEV);}

char vflag = 0, tflag = 0, bflag = 0, eflag = 0, dflag = 0,
     Kflag = 0, rflag = 0, lflag = 0, kflag = 0;

char YyTabFile[FNAMESIZE];
MAIN
{
	int sig;
	char *out = 0, *dir=0; 
	void catch(), child();
         
	for (sig=1; sig<=LASTSIG; sig++)
		if ((sig != SIGKILL) &&  (sig != SIGCLD))
				signal(sig,catch);
				 
	OPT
	ARG 'r':
		rflag = 1; 	
	ARG 'v':
		vflag = 1;	/* verbose flag				*/
	ARG 't':
		tflag = 1;	/* trace facilities for gen. parser	*/
	ARG 'b':		/* envoke output filter m2b		*/
		bflag = 1;
	ARG 'l':		/* cause gen. parser to load tables     */
		lflag = 1;
	ARG 'k':
		kflag = 1;	/* keep (do not delete) outputfiles     */
	ARG 'K':
		Kflag = 1;
	ARG 'D':
		dflag = 1;
	ARG 'e':
		eflag = 1;
	ARG 'W':
		lflag = 1;
		PARM
	 	dir = *argv;   /* tabfile to be put on this directory */
	        NEXTOPT
        ARG 'o': 
		PARM    	/* change output name */
		out = *argv;
		NEXTOPT
	OTHER
		USAGE;
	ENDOPT
	if (argc != 1)
		USAGE;

	if (!out)
		out = DEFNAME; 
	dflag && fprintf(stderr,"\tgenfilenames() ...\n");
	genfilenames(out,dir);
	dflag && fprintf(stderr,"\topenfiles() ...\n");
	ffopen(finput,*argv,"r");
	ffopen(flocal,FLOCAL,"w");
	ffopen(fdata,FDATA,"w");
	ffopen(ftemp,FTEMP,"w");
	if (lflag)
		ffopen(ftab,FTAB,"w");
	ffopen(faction,FACTION,"w");
	dflag && fprintf(stderr,"\tinitfiles() ...\n");
	initfiles(*argv,bflag); 
	mem = mem0;
	dflag && fprintf(stderr,"\tyyparse() ...\n");
	yyparse(); 
	ffclose(finput);
	ffclose(faction);
	tbitset	= NWORDS(ntokens);
	dflag && fprintf(stderr,"\tcpres() ...\n");
	cpres();
	dflag && fprintf(stderr,"\tcempty() ...\n");
	cempty();
	dflag && fprintf(stderr,"\tcpfir() ...\n");
	cpfir(); 
	dflag && fprintf(stderr,"\tstagen() ...\n");
        stagen();
	dflag && fprintf(stderr,"\toutput() ...\n");
	output();
	dflag && fprintf(stderr,"\tgo2out() ...\n");
	go2out();
	dflag && fprintf(stderr,"\thideprod() ...\n");
	hideprod();
	dflag && fprintf(stderr,"\tsummery() ...\n");
	summary();
	dflag && fprintf(stderr,"\tcallopt() ...\n");
	callopt();
	dflag && fprintf(stderr,"\tdefout() ...\n");
	defout();
	dflag && fprintf(stderr,"\tothers() ...\n");
	others();
	dflag && fprintf(stderr,"\texpect() ...\n");
	expect();
	ffclose(flocal);
	ffclose(fdata);
	if (lflag)
		ffclose(ftab);
	dflag && fprintf(stderr,"\tgenparser() ...\n");
	genparser();
	dflag && fprintf(stderr,"\tcloseupfile() ...\n");
	closeupfiles(bflag);
	dflag && fprintf(stderr,"\tdeltmpfiles() ...\n");
	deltmpfiles();
	dflag && fprintf(stderr,"\texit() ...\n");
	return EX_SUC;
}

genfilenames(name,dir)

char *name, *dir;

{
	if (strlen(name) > SIGNIFICANT)
	{	name[SIGNIFICANT] = '\0';
		fprintf(stderr,"warning: cut modname to %s\n",name);
	}
	strcpy(FM2,name);
	strcat(FM2,M2);
	strcpy(FDEF,name);
	strcat(FDEF,DEF);
	if (rflag)
		strcat(FDEF,"y");
	if (vflag)
	{	strcpy(FOUT,name);
		strcat(FOUT,VERBOSE);
	}
	if (lflag)
	{	
		strcpy(FTAB,name);
		strcat(FTAB,TAB);
		if (dir)
		{	strcpy(YyTabFile,dir);
		    	if (YyTabFile[strlen(YyTabFile)-1] != '/')
				strcat(YyTabFile,"/");
		   	strcat(YyTabFile,FTAB);
		}
	        else
		    strcpy(YyTabFile,FTAB);
	}
	strcpy(FLOCAL,name);
	strcat(FLOCAL,LOC);
	strcpy(FDATA,name);
	strcat(FDATA,DAT);
	strcpy(FTEMP,name);
	strcat(FTEMP,TMP);
	strcpy(FACTION,name);
	strcat(FACTION,ACT);
	strcpy(MODULE,name);
	if (access(".",02) == -1)
		error("cannot create files on current directory\n");
}

	
closeup()

{	ffclose(finput);
	if (bflag)
	{	pclose(fm2);
		pclose(fdef);
	}
	else
	{ 	ffclose(fm2);
		ffclose(fdef);    
	}
	ffclose(fparser);
	ffclose(ftemp);   
	ffclose(fdata);  
	ffclose(faction);
	ffclose(flocal);  
	if (lflag)
		ffclose(ftab);
	
}

static void catch(code)
int code;
	
{
	fprintf(stderr,"\n[interrupt: %d]\n\n",code); 
	closeup();          
	deltmpfiles();
	deloutfiles();
	exit(EX_IRP);
}

static void child()

{	fprintf(stderr,"\n[interrupt]: death of a child (may be m2b)");
	fprintf(stderr, "\n             check syntax of modula-2-text");
	fprintf(stderr," or try without -b option\n");
	closeup();
	deltmpfiles();
	exit(EX_IRP);
}

deltmpfiles()

{	if (!Kflag)
	{	ZAPFILE(FLOCAL);
		ZAPFILE(FTEMP); 
		ZAPFILE(FACTION);
		ZAPFILE(FDATA); 
	}
}
 
deloutfiles()

{	if (!kflag)
	{	ZAPFILE(FM2);
		ZAPFILE(FDEF);
		ZAPFILE(FOUT);
		if (lflag)
			ZAPFILE(FTAB);
	}
}
