/*======================================================================*
 *		        						* 
 * TITLE:	MYACC -- LALR(1) parser generator for modula-2	        *
 *									*
 * FILE:	parse.c - syntax analysis module 			*
 * 					        			* 
 * SUMMARY:   	yyparse() is myacc's parser. It checks the syntax of	*
 * 		the source file and fills global arrays containing 	*
 *		information about productions, precedence and 		*
 *		assosiatives.						*
 * 		Syntax errors are treated within this module.		*	
 *								        *
 * INCLUDES:	mdefs.h - system dependaant information and globals	*
 *		token.h - tokendefintions				*
 *		key.h   - table containing reserved words		*
 * 									*	
 * UPDATED:     July 5, 1987						*
 *								        *
 * AUTHOR: 	Werner Stanglow						*
 *								        *
 *======================================================================*/
    
#include "mdefs.h"	/* system dependant information and globals     */
#include "token.h" 	/* tokendefinitions 				*/
#include "search.h"	

/* External =========================================================== */
extern fatfl;		/* to mark errors fatal		(my1.c)		*/
extern char *strsave();	/* save a string 		(search.c)	*/
extern char *lltext;
extern gettok();
extern char *tokname;	/* name of token,		(lex.c)		*/
extern numbval;
extern char dflag;

/* Local definitions ================================================== */
#define D_MOD 01	/* flag, if set text is copied to fdef		*/
#define I_MOD 02	/* flag, if set text is copied to fdef		*/
#define CID 01
#define RES 02 
 
/* Variables (global to this module) ================================== */
static int  ntypes;	      	/* number of types defined		*/
static char *typeset[NTYPES]; 	/* pointers to type tags		*/
static int  start;  		/* start symbol				*/
static int casedef;		/* case defined ? 			*/
/*
static int keyword[127];*/	/* keywords				*/
static int keyno =0;
static int synerr = 0;
static int handle;
static int maxtoklen = 12;	/* maximum stringlength of tokennames	*/

/* -------------------------------------------------------------------- */
static char *cstash(s)		/* store s in a local memory and return */
char *s;			/* its address				*/
{	
	static char cnames[CNAMSZ];   	/* memory to store names	*/
	static char *cnamp = cnames;  	/* next free position in cnames */
	register char *tmp = cnamp; 		

	if ((cnamp+strlen(s)+1) > &cnames[CNAMSZ])
		error("too many characters in id's and literals" );
	while(*cnamp++ = *s++);
	return tmp;
}

/* -------------------------------------------------------------------- */
static defin(t,s)		/* define s to be a terminal (t==0) or 	*/
int t;				/* nonterminal symbol (t>0) 		*/
char *s;

{	register val;

	if (t)
	{	if ( ++nnonter >= NNONTERM)   
			error("too many nonterminals, limit %d", NNONTERM);
		nontrst[nnonter].name =	cstash(s);
		return(	NTBASE + nnonter );
	}
	/* must	be a token */
	if ( ++ntokens >= NTERMS )
		error("too many terminals, limit %d", NTERMS );
	tokset[ntokens].name = cstash(s);
	
	/* establish value for token */
	if (s[0] == ' ')
		if (s[2])
			error ("myacc internal: bad character format ");
		else
			val = s[1];
	else
	{	
		register x = strlen(s);
		val = extval++;
		maxtoklen = maxtoklen > x ? maxtoklen : x;
	}
	tokset[ntokens].value =	val;
	toklev[ntokens]	= 0;
	return ntokens;
}

/* -------------------------------------------------------------------- */
static synerror(fmt, a1 /*VARG */)
				/* put out syntax errors and skip input */ 
				/* until handle is read			*/
char *fmt;			/* - format element of fprintf()	*/
char *a1;			/* - argument list			*/

{	register t;

	fprintf(stderr,"[syntax error in line %d] ",lineno);
	fprintf(stderr,fmt,a1);
	putc('\n',stderr);
	synerr++;
	while ((t=gettok()) != ENDFILE)
	{
		switch (handle) {
		case RES:		
			if (isreserved(t))
				break;
			else 
				continue;
		case CID:
			if ( t==C_IDENTIFIER || t==';' ) 
				break;
			else 
				continue;
		default:
			error("myacc internal: undefined case label");
		}	
		break;
	}
	if (t == ENDFILE)
		error("cannot recover from syntax errors");
	return t;
}

/* -------------------------------------------------------------------- */
static gettypename()	/* check a '<' IDENTIFIER '>' typedefinition	*/
			/* return index of type, -1 in case of error 	*/

{	register i;   

	if ( (gettok() != IDENTIFIER) || (gettok() != '>') )
		return -1;
	for (i=1; i <=ntypes; ++i) 
		if (!strcmp(typeset[i],tokname))
			return i;
	typeset[++ntypes] = cstash(tokname);
	return ntypes;
} 

/* -------------------------------------------------------------------- */
static chfind(t,s)	/* look up index of grammar symbol s	 	*/
register t;		/* - nonzero means nonterminals, if 1 define 	*/
register char *s;	/*   s to be a nonterminal			*/

{ 	register i;

	if (s[0] == ' ')
		t = 0;
	TLOOP(i)
		if (!strcmp(s, tokset[i].name))
			return(	i );
	NTLOOP(i)
		if (!strcmp(s, nontrst[i].name)) 
			return(	i + NTBASE );
	/* cannot find name */
	if (t>1)
		error( "%s should have been defined earlier", s	);
	return (defin(t,s));
}

/* -------------------------------------------------------------------- */
static cpycase()	/* output of casevariant record appropiate to	*/
			/* modula 2					*/
{	register t, label = 0;

	fprintf( fdef,"(* line: %d *)\n",lineno);
	fprintf( fdef,"TYPE\n\tYYSTYPE = \n\t\tRECORD\n\t\t\tCASE : CARDINAL OF\n"); 
        for(;;)
	{ 
		switch (t=gettok())
		{
			case C_IDENTIFIER:
				fprintf(fdef,"\t\t\t| %2d : ",++label);
			case IDENTIFIER:
			case ';':
				break;
			case END:
				fprintf(fdef,"\t\t\tEND\n\t\tEND;\n");
				return END; 
				break;
			default:
				return t;
		}
		fprintf(fdef,"%s",lltext); 
		continue; 
	}
}

/* -------------------------------------------------------------------- */
static fdtype(t)		/* determine the type of a symbol 	*/
register t;
{
	register v;
	if ( t >= NTBASE )
		v = nontrst[t-NTBASE].tvalue;
	else
		v = TYPE( toklev[t] );
	if ( v <= 0 )
	{	synerror( "must specify type for %s", (t >= NTBASE) ? nontrst[t-NTBASE].name
		    :	 tokset[t].name	);
		return 1;
	}
	return(	v );
}


/* ==================================================================== */
yyparse()		/* parses file stream finput			*/

{	int c, i=0, j, * p, lev=0, t, tempty, ty = 0, iskey = 0;
	char actname[8];
	char startdef = 0;

/* fashion first rule */
	defin(0, "$end");	
	extval = 0400;		
	defin(0, "error");
	defin(1, "$accept");

        handle = RES;

/* check interface description */
	for(t=gettok();(t==IMPORTS) || (t==INIT);)
	{	
		switch(t)
		{	case IMPORTS: 
				t = cpytext(D_MOD|I_MOD);
				continue;
			case INIT:
				if (gettok() != IDENTIFIER)
				{	t = synerror(
						"bad syntax near %s",tokname);
					continue;
				}
				strcpy(initproc,tokname);
				strcpy(initmodule,tokname);
				if ( (t=gettok()) != '.')
				{	*initmodule = '\0';
					continue;
				}
				else if ((t=gettok()) != IDENTIFIER)
				{	t = synerror(
						"bad syntax near %s",tokname);
					continue;
				}
				else
				{	
					strcat(initproc,".");
					strcat(initproc,tokname);
					t = gettok(); 
					continue;
				}
			default:
				break;
		}
		break;
	}
	genimports();
/* parse declaration part */
	while (t != MARK)
	{
		switch (t)
		{	
			case ENDFILE:
				t = synerror("Unexpected EOF before '%%%%'");	
				continue;
			case ERRORTOK:
				t = synerror("illegal reserved word: %%%s",tokname);
				continue;
			case START:
				if (startdef)	
				{	t = synerror ("redeclaration of start symbol");
					continue;
				}
				
				if ( (t	= gettok()) != IDENTIFIER )
				{	t = synerror( "bad %%start construction" );
					continue;
				}
				start =	chfind(1, tokname);
				startdef = 1;    
				t = gettok();
				continue;
			case TYPEDEF:
				if (gettok() != '<') 
				{	t = synerror ("'<' expected");
					continue;
				}
				if ((ty = gettypename()) < 0)
				{	t = synerror("bad type definition");
					continue;
				}
				while ((t = gettok()) == IDENTIFIER)
				{	if ( (t	= chfind( 1, tokname ) ) <
					    NTBASE ) 
					{	j = TYPE( toklev[t] );
						if ( j != 0 && j != ty )
							synerror( "type redeclaration of token %s", tokset[t].name	);
						else
							SETTYPE( toklev[t],
							     ty);
					} 
					else
					{	j = nontrst[t-NTBASE].tvalue;
						if ( j != 0 && j != ty ) 
							synerror( "type redeclaration of nonterminal %s", nontrst[t-NTBASE].name);
						else
							nontrst[t-NTBASE].tvalue = ty;
					}
				}
				continue; 
			case DEFS:
				t = cpytext(D_MOD);
				continue;
			case CASE:
				if (casedef)
				{	 t = synerror("%%case already defined");
					 continue;
				}
				if ((t = cpycase()) != END)
				{	t = synerror("%%end expected",t);
					continue;
				}
				casedef = 1;
				t = gettok();
				continue;
			case END:
				t = synerror("%%end does not match");
				continue;
			case RCURLY: 
				t = synerror("%%} does not match");
				continue;
			case LEX:
			case ERROR:
			case INIT:
			case IMPORTS:
				t = synerror ("%%%s should have been used erlier",
					tokname);
				continue;
			case KEY:
				iskey = 1;
				t = TOKEN;
				continue; /* i.e goto TOKEN: */
			case LEFT:
			case BINARY:
			case RIGHT:
				++i;
			case TOKEN:
				lev = t	- TOKEN;	 /* nonzero means new prec. and	assoc. */
				ty = 0;
				t = gettok();
				if (t == '<' )
				{	/* there is a type defined */
					if ((ty = gettypename()) < 0)
					{	t = synerror("bad type definition");
						continue;
					}
					t = gettok();
				}
				while (t == IDENTIFIER)
				{	j = chfind(0, tokname);
					if (iskey)
					        tokset[j].iskey = 1;
						/*keyword[keyno++] = j;*/
					if (lev)
					{	if ( ASSOC(toklev[j]) )
							synerror( "redeclaration of precedence of%s", tokname );
						SETASC(toklev[j], lev);
						SETPLEV(toklev[j], i);
					}
					if (ty)
					{	if ( TYPE(toklev[j]) )
							synerror( "redeclaration of type of %s", tokname );
						SETTYPE(toklev[j], ty);
					}
					t = gettok();
				}
				iskey = 0;
				continue;
			case LCURLY:
				if((t = cpytext(I_MOD)) != RCURLY)
					synerror("%%} expected");
				t = gettok();
				continue;
			default:
				t = synerror("illegal token or character (octal %o)",
				t);
				continue;

		}
	}


/* init mem and prdptr */
	prdptr[0] = mem;
	*mem++ = NTBASE;
	*mem++ = start;	 /* if start is	0, we will overwrite  */
			 /* with  the lhs of the first rule   */  
	*mem++ = 1;
	*mem++ = 0;
	prdptr[1] = mem;
        handle = CID;
	if ( (t=gettok()) != C_IDENTIFIER )
		t = synerror( "bad syntax on first rule" );
	if ( !start )
		prdptr[0][1] = chfind(1, tokname);

/* now read the rules */
	while (	t != MARK && t != ENDFILE ) 
 	{
	 	if ( t == C_IDENTIFIER )
		{	*mem = chfind(1, tokname);
			if ( *mem < NTBASE )
			{	t = synerror( "token illegal on LHS of grammar rule" );
				continue;
			}
			++mem;
		}
	        else if (t == '|')	
			*mem++ = *prdptr[nprod-1];
		else
		{	t = synerror("illegal rule: expected '|' or ';'");
			continue;
		}

		/* read	rule body */
		t = gettok();

more_rule:      dflag && fprintf(stderr,"more_rule\n");
		while (	t == IDENTIFIER	)
		{	
			dflag && fprintf(stderr,"chfind(1,%s)\n",tokname);
			*mem = chfind(1, tokname);
			dflag && fprintf(stderr,"chfind(1,%s)=%u\n",tokname,*mem);
			if ( *mem < NTBASE )
				levprd[nprod] =	toklev[*mem];
		        dflag && fprintf(stderr,"beyond if *mem < toklev\n");
			++mem;
			t = gettok();
		        dflag && fprintf(stderr,"byond t = gettok in while(=%u)\n",t);
		}
		dflag && fprintf(stderr,"beyond while\n");
		if ( t == PREC ) 
		{	if ( gettok() != IDENTIFIER)
			{	t = synerror( "illegal %%prec syntax" );
				continue;
			}
			j = chfind(2, tokname);
			if ( j >= NTBASE)
			{	synerror("nonterminal %s illegal after %%prec",
				     nontrst[j-NTBASE].name);
				continue;
			}
			levprd[nprod] =	toklev[j];
			t = gettok();
		}
		dflag && fprintf(stderr,"beyond t == PREC\n");
		if (t == '{') /* } */	
		{	levprd[nprod] |= ACTFLAG;
			fprintf( faction, "\n| %3d:", nprod );
			cpyaction(mem - prdptr[nprod] - 1);
			dflag && fprintf(stderr,"\t\tback from cpyaction (%u)\n", lineno);
			if ( (t	= gettok()) == IDENTIFIER )
			{ 	/* action within rule... */
				int countDbg = 0;
				sprintf( actname, "$$%d", nprod	);
			dflag && fprintf(stderr,"chfind(1,%s)\n",actname);
				j = chfind(1, actname);	 /* make it a nonterminal */
			dflag && fprintf(stderr,"chfind(1,%s)=%u\n",actname,j);
				/* the current rule will become	rule number nprod+1 */
				/* move	the contents down, and make room for the null */
			dflag && fprintf(stderr,"at %u\n",++countDbg);
				for ( p	= mem; p >= prdptr[nprod]; --p )
					p[2] = *p;
			dflag && fprintf(stderr,"at %u\n",++countDbg);
				mem += 2;
				/* enter null production for action */
			dflag && fprintf(stderr,"at %u\n",++countDbg);
				p = prdptr[nprod];
				*p++ = j;
				*p++ = -nprod;

				/* update the production information */
			dflag && fprintf(stderr,"at %u\n",++countDbg);
				levprd[nprod+1]	= levprd[nprod]	& ~ACTFLAG;
			dflag && fprintf(stderr,"at %u\n",++countDbg);
				levprd[nprod] =	ACTFLAG;
			dflag && fprintf(stderr,"at %u\n",++countDbg);

				if ( ++nprod >=	NPROD )
					error( "more than %d rules", NPROD );
			dflag && fprintf(stderr,"at %u\n",++countDbg);
				prdptr[nprod] =	p;

				/* make	the action appear in the original rule */
				*mem++ = j;

				/* get some more of the	rule */

			dflag && fprintf(stderr,"at %u\n",++countDbg);
				goto more_rule;
			}
		}
		dflag && fprintf(stderr," t == '{'\n");
		*mem++ = -nprod;

		/* check that default action is	reasonable */

		if ( ntypes && !(levprd[nprod] & ACTFLAG) && nontrst[*prdptr[nprod]-NTBASE].tvalue ) 
		{	 /* no explicit action, LHS has value */
			/*01*/
			tempty = prdptr[nprod][1];
			if ( tempty < 0	)
			{	t = synerror( "must return a value, since LHS has a type" );
				continue;
			}
			else if	( tempty >= NTBASE )
				tempty = nontrst[tempty-NTBASE].tvalue;
			else
				tempty = TYPE( toklev[tempty] );
			if ( tempty != nontrst[*prdptr[nprod]-NTBASE].tvalue ) 
			{ 	t = synerror( "default action causes potential type clash" );
				continue;
			}

		}
		if ( ++nprod >=	NPROD )
			error( "more than %d rules", NPROD );
		prdptr[nprod] =	mem;
		levprd[nprod] =	0;
		if (t == ';')
			t = gettok();
	}
	/* end of all rules */

	if ( t == MARK )
		cpytext(I_MOD);
	if (!casedef)
		fprintf(fdef,"\nTYPE\n\tYYSTYPE = INTEGER;\n");
        if (synerr)
	{
		fatfl = 0;
		error("--- %d syntax errors detected",synerr);
		fatfl = 1;
	}
}

/* -------------------------------------------------------------------- */
static cpytext(mode)	/* Copies text from file finput to fdef and/or  */
			/* fm2 (depends on mode), returns next keyword  */
unsigned mode;

{	register t;
	
	if (mode & D_MOD)
		fprintf(fdef,"\n(* line: %d *)", lineno );
	if (mode & I_MOD)
		fprintf(fm2,"\n(* line: %d *)", lineno );
	while (((t = gettok()) != ENDFILE) && (!isreserved(t)))
	{
		if (mode & D_MOD)
			fputs(lltext,fdef);
		if (mode & I_MOD)
			fputs(lltext,fm2);
	}
	return t;
}

/* -------------------------------------------------------------------- */
static cpyaction(offset)	/* Copies text to the matching '}' and 	*/
				/* substitutes the $x-constructions.  	*/
register offset;		/* Attributes 1..offset may be used.	*/	
				/* Returns next keyword.		*/

{       register brac, c;
	int begin;

	fprintf( faction, "\t(* line: %d *)\n", lineno );
	begin = lineno;
	brac = 1;

	while (brac)
	{	
		switch (c = gettok())
		{	case ENDFILE:
		error ("action starting at line %d does not terminate",begin);
			case '{':
				brac++;
				break;
			case '}':
				if (--brac)
					break;
				else
					return '}';
			case '$':
				outattribute(offset);
				continue;
		}
		fputs(lltext,faction);
	}
}

/* -------------------------------------------------------------------- */
static outattribute(offset)  
			/* output of an attribute used within an action	*/
int offset;		/* (see above)					*/

{	register c, s, j, tok;

	s = 1;
	tok = -1;
	if ((c = gettok()) == '<')
	{	tok = gettypename();
		if (tok < 0) 
		{	c = synerror("bad type in $<..>-construct");
			return;
		}
		c = gettok();  
	}
	if (c == '$')	
	{	fprintf(faction, "yyval");
		if (ntypes) 
		{	if (tok<0)
				tok = fdtype(*prdptr[nprod] );
			fprintf( faction, ".%s", typeset[tok] );
		}
		return; 
	}
	if ( c == '-' )	
	{	s = -s;
		c = gettok();
	}
	if (c == NUMBER)
		j = numbval*s - offset;	
	else 
		synerror("number or '$' expected in $.-construction");
	if (j > 0) 
		synerror("Illegal use of $%d",j+offset);
	fprintf( faction, "yyv[yypvt-(%d)]",-j);
	if (ntypes) 
	{	if (j+offset <= 0 && tok < 0)
			synerror("must specify type of $%d", j + offset	);
		if ( tok < 0 )
			tok = fdtype( prdptr[nprod][j+offset] );
		fprintf( faction, ".%s", typeset[tok] );
	}
}
